#ifndef _DRAWWINDOW_HH_
#define _DRAWWINDOW_HH_

#include <gtk/gtk.h>
#include <string>
using std::string;

class DrawWindow
{
public:
    inline DrawWindow(unsigned width, unsigned height,
                      void (*DrawFunction)(DrawWindow&),
                      int& argc, char**& argv,
                      const string& bgcolor="black");
    /* width, height : Initial width and height of the window.
       DrawFunction : Pointer to function that draws the picture. Example:
                      void Draw(DrawWindow& draw)
                      { draw.Line(0,0, draw.Width(),draw.Height());
                      }
       argc, argv : Command line parameters given to main().
       bgcolor : Background color of the drawing window.
     */

    inline void OnClick(void (*OnClickFunction)
                        (DrawWindow&, unsigned x, unsigned y,
                         unsigned button));
    /* When the drawing window is clicked, the given function is called with
       the click coordinates.
     */

    inline void Done();
    /* This is called after creating and setting up the class. The class
       starts waiting for events.
     */

    inline void DrawingColor(const char* color);
    inline void DrawingColor(unsigned red, unsigned green, unsigned blue);
    inline void DrawingColor(unsigned color);
    /* Set the drawing color for next drawing commands.
       - The first one is in the form, for example, "#FF8800" or "yellow".
       - The second one takes three color components; values 0 - 65535.
       - The third one takes one unsigned value; the lower 8 bits are the red
         component, the next 8 bits the green and the next 8 bits the blue.
     */

    inline unsigned Width() { return width; }
    inline unsigned Height() { return height; }
    /* Width and height of the current window.
     */

    inline void Point(unsigned x, unsigned y);
    inline void Point_immed(unsigned x, unsigned y);
    inline void Line(unsigned x1, unsigned y1, unsigned x2, unsigned y2);
    inline void Circle(unsigned x, unsigned y, unsigned radius);
    /* Some drawing commands.
     */

    inline void RefreshArea(unsigned x, unsigned y,
                            unsigned width, unsigned height);
    /* The drawing commands draw on a pixmap which is showed after the
       call to 'DrawFunction' has ended (for speed reasons).
       If, however, you want to show the picture while drawing it, you
       can call this method and tell it to show the given area of the
       window immediately. If, for example, you are drawing something
       pixel by pixel (eg. a Mandelbrot set) and want to show the result
       after each line, you can call this method after each line and tell
       it to show the current line. The call would be in this case something
       like: draw.RefreshArea(0,currentline, draw.Width(),1);
     */

    inline void Clear();

//============================================================================
//============================================================================
private:
    struct GtkWidget_C: public GtkWidget
    {
        DrawWindow* instance;

        inline GtkWidget_C& operator=(GtkWidget* widget)
        {
            *((GtkWidget*)this) = *widget;
            return *this;
        }
    };

    GtkWidget *mainwindow;
    GdkGC* DrawingGC;
    GdkColormap* ColorMap;
    GtkWidget_C darea;
    GdkPixmap *pixmap;
    string bgcolor;
    void (*DrawFunction)(DrawWindow&);
    void (*OnClickFunction)(DrawWindow&, unsigned, unsigned, unsigned);

    unsigned width, height;

    static gint configure_event(GtkWidget_C *widget, GdkEventConfigure*);
    static gint expose_event(GtkWidget_C *widget, GdkEventExpose *event);
    static gint button_press_event(GtkWidget_C *widget, GdkEventButton *event);
};

inline void DrawWindow::DrawingColor(const char* color)
{
    GdkColor col;
    gdk_color_parse(color,&col);
    gdk_color_alloc(ColorMap,&col);
    gdk_gc_set_foreground(DrawingGC, &col);
}
inline void DrawWindow::DrawingColor(
    unsigned red, unsigned green, unsigned blue)
{
    GdkColor col={0,red,green,blue};
    gdk_color_alloc(ColorMap,&col);
    gdk_gc_set_foreground(DrawingGC, &col);
}
inline void DrawWindow::DrawingColor(unsigned color)
{
    GdkColor col={0,(color&255)<<8,color&(255<<8),(color>>8)&(255<<8)};
    gdk_color_alloc(ColorMap,&col);
    gdk_gc_set_foreground(DrawingGC, &col);
}

inline void DrawWindow::Point(unsigned x, unsigned y)
{
    gdk_draw_point(pixmap, DrawingGC, x, y);
}

inline void DrawWindow::Point_immed(unsigned x, unsigned y)
{
    gdk_draw_point(pixmap, DrawingGC, x, y);
    gdk_draw_point(darea.window, DrawingGC, x, y);
}

inline void DrawWindow::Line(
    unsigned x1, unsigned y1, unsigned x2, unsigned y2)
{
    gdk_draw_line(pixmap, DrawingGC, x1, y1, x2, y2);
}
inline void DrawWindow::Circle(unsigned x, unsigned y, unsigned radius)
{
    gdk_draw_arc(pixmap, DrawingGC,
                 FALSE, x-radius, y-radius, 2*radius, 2*radius, 0, 24000);
}

inline DrawWindow::DrawWindow(unsigned width, unsigned height,
                              void (*DrawFunction)(DrawWindow&),
                              int& argc, char**& argv,
                              const string& bgcolor):
    pixmap(0), bgcolor(bgcolor), DrawFunction(DrawFunction), OnClickFunction(0)
{
    gtk_init (&argc, &argv);
    mainwindow = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_window_set_policy(GTK_WINDOW(mainwindow), TRUE, TRUE, FALSE);
    gtk_signal_connect(GTK_OBJECT(mainwindow), "destroy",
                       GTK_SIGNAL_FUNC(gtk_main_quit), NULL);
    gtk_widget_set_usize(GTK_WIDGET(mainwindow), width, height);

    darea = gtk_drawing_area_new();
    darea.instance = this;
    gtk_container_add(GTK_CONTAINER(mainwindow), &darea);

    gtk_widget_show(mainwindow);

    DrawingGC=gdk_gc_new(mainwindow->window);
    gdk_gc_copy(DrawingGC, mainwindow->style->fg_gc[0]);
    ColorMap=gdk_colormap_get_system();

    gtk_signal_connect (GTK_OBJECT(&darea),"expose_event",
                        (GtkSignalFunc) expose_event, NULL);
    gtk_signal_connect (GTK_OBJECT(&darea),"configure_event",
                        (GtkSignalFunc) configure_event, NULL);
    gtk_signal_connect (GTK_OBJECT (&darea), "button_press_event",
                        (GtkSignalFunc) button_press_event, NULL);

    gtk_widget_set_events (&darea, GDK_EXPOSURE_MASK
                           | GDK_LEAVE_NOTIFY_MASK
                           | GDK_BUTTON_PRESS_MASK);

    gtk_widget_show(&darea);
}

inline void DrawWindow::OnClick(void (*OnClickFunction)
                                (DrawWindow&, unsigned x, unsigned y,
                                 unsigned button))
{
    this->OnClickFunction = OnClickFunction;
}

inline void DrawWindow::Done() { gtk_main(); }

inline void DrawWindow::RefreshArea(unsigned x, unsigned y,
                                    unsigned width, unsigned height)
{
    gdk_draw_pixmap(darea.window,
                    darea.style->fg_gc[GTK_WIDGET_STATE (&darea)],
                    pixmap,
                    x, y, x, y, width, height);
}

inline void DrawWindow::Clear()
{
    DrawingColor(bgcolor.c_str());
    gdk_draw_rectangle (pixmap,
                        DrawingGC,
                        TRUE,
                        0, 0, Width(), Height());

    //RefreshArea(0, 0, Width(), Height());
}

gint DrawWindow::configure_event(GtkWidget_C *widget, GdkEventConfigure*)
{
    if (widget->instance->pixmap)
        gdk_pixmap_unref(widget->instance->pixmap);

    widget->instance->pixmap = gdk_pixmap_new(widget->window,
                                              widget->allocation.width,
                                              widget->allocation.height,
                                              -1);
    widget->instance->DrawingColor(widget->instance->bgcolor.c_str());
    gdk_draw_rectangle (widget->instance->pixmap,
                        widget->instance->DrawingGC,
                        TRUE,
                        0, 0,
                        widget->allocation.width,
                        widget->allocation.height);

    widget->instance->RefreshArea(0, 0,
                                  widget->allocation.width,
                                  widget->allocation.height);

    widget->instance->width = widget->allocation.width;
    widget->instance->height = widget->allocation.height;
    widget->instance->DrawFunction(*widget->instance);

    return TRUE;
}

gint DrawWindow::expose_event(GtkWidget_C *widget, GdkEventExpose *event)
{
    gdk_draw_pixmap(widget->window,
                    widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
                    widget->instance->pixmap,
                    event->area.x, event->area.y,
                    event->area.x, event->area.y,
                    event->area.width, event->area.height);

    return FALSE;
}

gint DrawWindow::button_press_event(GtkWidget_C *widget,
                                    GdkEventButton *event)
{
    if(widget->instance->OnClickFunction)
        widget->instance->OnClickFunction(*widget->instance,
                                          (unsigned)event->x,
                                          (unsigned)event->y,
                                          (unsigned)event->button);
    return FALSE;
}

#endif
